from woke.cli.detect import run_detect as detector

from .api import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
)
from .balance_state_var import UnsafeAddressBalanceUseDetector
from .bug_empty_byte_array_copy import BugEmptyByteArrayCopyDetector
from .call_options_not_called import FunctionCallOptionsNotCalledDetector
from .missing_return import MissingReturnDetector
from .msg_value_nonpayable_function import MsgValueNonpayableFunctionDetector
from .overflow_calldata_tuple_reencoding_bug import (
    OverflowCalldataTupleReencodingBugDetector,
)
from .proxy_contract_selector_clashes import ProxyContractSelectorClashDetector
from .reentrancy import ReentrancyDetector
from .unchecked_return_value import UncheckedReturnValueDetector
from .unsafe_delegatecall import UnsafeDelegatecallDetector
from .unsafe_selfdestruct import UnsafeSelfdestructDetector
from .unsafe_tx_origin import UnsafeTxOriginDetector
from .unused_contract import UnusedContractDetector

from . import axelar  # isort:skip
