import lazy_import

lazy_import.lazy_module("networkx")
lazy_import.lazy_module("woke.ir")
lazy_import.lazy_module("woke.ir.enums")
lazy_import.lazy_module("woke.ir.types")
lazy_import.lazy_module("woke.analysis")
lazy_import.lazy_module("woke.analysis.cfg")
lazy_import.lazy_module("woke.analysis.ownable")
lazy_import.lazy_module("woke.analysis.utils")
